"""
Simplified statistics
"""
from .stat import unv,usd,unv_lambda, poisson_dist,normalize,novar_mean,mean,noisy

__all__ = ['unv',
'usd',
'unv_lambda', 'poisson_dist','normalize','novar_mean',
'mean','noisy'
]


import numpy as np

def times(str,int):
    return str.join(["" for i in range(0,int+1)])

def has(key, dict):
    """
        Checks if the key is in the dict and not None.
    """
    return key in dict and not dict[key] is None

def true(key, dict):
    """
        Checks if the key is in the dict and not None and True.
    """
    return has(key,dict) and dict[key]


def find_nearest_index(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx
def find_nearest(array, value):
    array[find_nearest_index(array,value)]



import pathlib
import glob
import pathlib
#from io import StringIO
from smpl import debug
import numpy as np
import os
import sys
from smpl.io import mkdirs

def gf(i):
    """
    Scientific format with ``i`` digits.
    """
    return "{0:." + str(i) + "g}"

def si(s,u="",fmt="{}"):
    """
    Get number with uncertainty and unit in ``si`` format for latex.

    Parameters
    ==========
    s : ufloat
        number to be returned in a latex compatible format
    u : str
        unit of that number
    fmt : str
        format string for the numbers

    Returns
    =======
    sistr : str
        latex SI string of the number with it's uncertainty and unit.
    """
    return "\\SI{%s}{%s}"%((fmt.format(s)).replace("/","").replace("(","").replace(")",""),u)

def si_line(a,skip = 0,fmt="{}"):
    """
    Get array ``a`` in the format of a line of a latex table.
    """
    return si_tab(np.transpose([[t] for t in a]),skip,fmt)
def si_ttab(tab,skip=0, fmt="{}"):
    """
    Transposed ``si_tab``.
    """
    return si_tab(np.transpose(tab),skip,fmt)
def si_tab(tab,skip=0, fmt="{}"):
    """
    Get arrays of (uncertainty) numbers in  a latex table compatible form.

    Parameters
    ==========
    tab : array_like
        Array containing the values of the table
    skip : number
        Skip this many table lines
    fmt : str
        format string for the numbers

    Returns
    =======
    tabstr : str
        table latex string
    """
    #mkdirs(fn)
    #file = open(fn,"w")
    s = ""
    for i in range(len(tab)):
        for j in range(len(tab[i])):
            if(j!=0):
                s += "&"
            if(j>=skip):
                s+=si(tab[i][j],fmt=fmt)
            else:
                s+="%s"%(tab[i][j])
        s += "\\\\\n"
    return s

def out_si(fn,s,u="",fmt="{}"):
    mkdirs(fn)
    file = open(fn,"w")
    file.write(si(s,u,fmt))
    print(fn,": ", fmt.format(s), u)
    file.close()
def out(fn,s):
    mkdirs(fn)
    file = open(fn,"w")
    file.write(("%s"%(s)).replace("/",""))
    print(fn,": ", "%s"%(s))
    file.close()


def out_si_line(fn,tab,skip=0):
    out_si_tab(fn,np.transpose([[t] for t in tab]),skip)
def out_si_tab(fn, tab,skip=0, fmt="{}"):
    mkdirs(fn)
    file = open(fn,"w")
    for i in range(len(tab)):
        for j in range(len(tab[i])):
            if(j!=0):
                file.write(pr("&",nnl=True))
            if(j>=skip):
                file.write(pr(si(tab[i][j],fmt=fmt),nnl=True))
            else:
                file.write(pr("%s"%(tab[i][j]),nnl=True))
        file.write(pr("\\\\\n",nnl=True))
    file.close()

def dump_vars(fd):
    key = ""
    gl = globals()
    for key in gl:
        smart_out(fd + key,gl[key])
    #print(globals())
def iter(a):
    return zip(range(len(a)),a)

def smart_out(fn,x):
    '''TODO'''
    out_si(fn,x)
    '''if isinstance(x,list):
        out_si_tab(fn,x)
    elif isinstance(x,types.FunctionType):
        out(fn,inspect.getsourcelines(x)[0])
    else:
        out_si(fn,x)
    '''

def frange(x, y, jump):
  while x < y:
    yield x
    x += jump


def pn(a,nnl=False):
    gl = globals()
    for key in gl:
        if(gl[key]==a):
            print(key)
    if nnl:
        print("%s=%s"%(a.__name__,a),end='')    
    else:
        print("%s=%s"%(a.__name__,a))
    return a
def pr(a,nnl=False):
    if nnl:
        print(a,end='')    
    else:
        print(a)
    return a




from inspect import currentframe,getsourcelines, getsource
import os
import numpy as np

DEBUG_LEVEL=-1
# DEBUG_LEVEL=-1 no debug
# DEBUG_LEVEL= 0 default level
# DEBUG_LEVEL= 1,2,3 higher levels, more debug

# print prefix
DEBUG_PRE="DBG"

# remove folders of files for print

# filename strings with .py (with folder if SPLIT_FILENAME==False)
WHITE_LIST_FILES=[]
BLACK_LIST_FILES=[]

active=debug=on=DEBUG_LEVEL>=0
# count debug events by file+line
count_times = {}
cur_table_line = {}
def get_frame(_back=0):
    cf = currentframe()
    for i in range(_back+1):
        cf = cf.f_back
    return cf
def once(_back=0):
    """
    Returns true only one time
    """
    return times(1,_back+1)
def times(t=1,_back=0):
   line,fname = get_line_number_file(_back+1) 
   inc_count(line,fname)
   return check_count(line,fname,t)

def get_line_src(_back=0):
    cf = get_frame(_back+1)
    srcline = getsource(cf).split("\n")[cf.f_lineno-cf.f_code.co_firstlineno].strip()
    return srcline

def get_line_number_file(split = True,_back=0):
    '''
    Gets the current filename and the current linenumber within it.

    Parameters
    ==========
    split : bool
        Indicates whenever the folders above of the file should be included in the returned filename.
    _back : int
        Number of stack/frames to go back.

    Returns
    =======
    filenumber : int
        First element in the return array
    filename : str
        Second element in the return array
    '''
    cf = get_frame(_back+1)
    fname = cf.f_code.co_filename
    if split:
        fname = cf.f_code.co_filename.split("/")[-1]
    return cf.f_lineno,fname

def get_line_number(_back=0):
    return get_line_number_file(_back+1)[0]

def line(msg_,tag="",level=0,times=-1,_back=0):
    msg(msg_,tag=tag,level=level,times=times,line_=True,_back=_back+1)
# only once
def line1(msg_,tag="",level=0,times=-1,_back=0):
    msg1(msg_,tag=tag,level=level,times=times,line_=True,_back=_back+1)

# counting functions
def get_count(line,fname):
    return count_times[fname+str(line)]

def inc_count(line,fname):
    if fname+str(line) in count_times:
        count_times[fname+str(line)]+=1
    else:
        count_times[fname+str(line)]=1

def check_count(line,fname,t):
    if t >= get_count(line,fname) or t ==-1:
        if(not fname in BLACK_LIST_FILES and (len(WHITE_LIST_FILES)==0 or fname in WHITE_LIST_FILES)):
            return True
    return False


# _line enables printint src line
# t stands for times
def msg(msg,tag="",level=0,times=-1,line_=False,_back=0):
    """
    Prints the message ``msg`` if level > debug_level
    """
    if(level<=DEBUG_LEVEL):
        line,fname = get_line_number_file(_back+1)
        src=""
        if line_ == True:
            src = get_line_src(_back+1)
            src = "(" + '('.join(src.split("(")[1:]) + " = "
        inc_count(line,fname)
        if(check_count(line,fname,times)):
            print(DEBUG_PRE + ":" + tag + ":" +fname + ":"+ str(line)  + ": "  + src + str(msg))
    return msg

def file(key,value,level=0,times=-1,seperator=";",_print=True,_back=0,filename="debug.csv"):
    """
    Prints the message ``msg`` if level > debug_level to file ``filename``
    """
    if(level<=DEBUG_LEVEL):
        line,fname = get_line_number_file(_back+1)
        inc_count(line,fname)
        if(check_count(line,fname,times)):
            f=open(filename,"a+")
            if(_print):
                tag = "debug.file"
                print(DEBUG_PRE + ":" + tag + ":" +fname + ":"+ str(line)  + ": " + key + seperator + value)
            f.write( key + seperator + value + "\n")
            f.close()
    return value 

def table_flush_header(filename = "debug_table.csv",seperator=";"):
    """
    Saves the current keys from :func:`table` to ``filename``
    """
    f=open(filename,"a+")
    for key in sorted(cur_table_line):
        f.write(key + seperator)
    f.write('\n')
    f.close()


def table_flush_line(filename = "debug_table.csv",seperator=";"):
    """
    Saves the current values from :func:`table` to ``filename``
    """
    f=open(filename,"a+")
    itt = iter(cur_table_line)
    ok = False
    while  not ok:
        cur = next(itt)
        ok = isinstance(cur_table_line[cur],np.ndarray)
    dim = len(cur_table_line[cur])
    for i in range(dim):
        for key in sorted(cur_table_line):
            if isinstance(cur_table_line[key],np.ndarray):
                f.write("%.30e"%cur_table_line[key][i] + seperator)
            else:
                f.write("%.30e"%cur_table_line[key] + seperator)
        f.write('\n')
    f.close()


def table(key,value,level=0,times=-1,seperator=";",_print=False,_back=0,filename="debug_table.csv"):
    """
    Saves ``key``:``value`` in ``filename``.
    """
    if(level<=DEBUG_LEVEL):
        line,fname = get_line_number_file(_back+1)
        inc_count(line,fname)
        if(check_count(line,fname,times)):
            if isinstance(value,np.ndarray):
                cur_table_line[key] = value.copy()
            else:
                cur_table_line[key] = value
    return value 



# only once
def msg1(_msg,tag="",level=0,times=1,line_=False,_back=0,**kwargs):
    """
    Just like :func:`msg` but ``times`` set to 1.
    """
    return msg(_msg,level=level,tag=tag,times=times,line_=line,_back=_back+1,**kwargs)
def file1(_key,_value,level=0,times=1,_back=0,**kwargs):
    """
    Just like :func:`file` but ``times`` set to 1.
    """
    return file(_key,_value,level=level,times=1,_back=_back+1,**kwargs)

# resets counts
def reset_times():
    count_times = {}
if os.path.exists("debug.csv"):
    os.remove("debug.csv")
if os.path.exists("debug_table.csv"):
    os.remove("debug_table.csv")



"""
A list of various functions
"""
from .functions import const, lorentz,linear,line,cos_abs,gauss,quadratic,exponential
__all__ = [
    'const', 
    'lorentz',
    'linear',
    'line',
    'cos_abs',
    'gauss',
    'quadratic',
    'exponential'
    ]


"""
Simplified python debuging
"""

from .debug import msg,once,msg1,table,file,file1,table_flush_header, table_flush_line,get_line_number_file

__all__ = [
    'msg', 
    'once', 
    'msg1', 
    'table', 
    'file', 
    'file1', 
    'table_flush_header',
    'table_flush_line',
    'get_line_number_file',
]


from smpl import util
params = 0
name = 1

def _append(txt):
    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        target.__doc__ += txt
        return target
    return wrapper

def append(txt):
    """
    TODO split cases doc/txt/plot?
    """
    return None

def append_str(txt):
    return _append(txt)

def append_plot(*args,xmin=-5,xmax=5):
    """
    Append a plot to a function.
    """
    #return _append("\n\n\t.. plot::\n\t\t:include-source:\n\n\t\t>>> from " + target.__module__ + " import " +target.__name__ + "\n\t\t>>> from smpl import plot\n\t\t>>> plot.function("+ target.__name__ + "," + ','.join([str(a) for a in args]) + ",xmin="+str(xmin) + ",xmax=" + str(xmax)+")")

    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        target.__doc__ += "\n\n\t.. plot::\n\t\t:include-source:\n\n\t\t>>> from " + target.__module__ + " import " +target.__name__ + "\n\t\t>>> from smpl import plot\n\t\t>>> plot.function("+ target.__name__ + "," + ','.join([str(a) for a in args]) + ",xmin="+str(xmin) + ",xmax=" + str(xmax)+")"
        #print(target.__doc__)
        return target
    return wrapper

def append_doc(original):
    """
    Append doc string of ``original`` to ``target`` object.

    Paramerters
    ===========
    original : ``class`` or ``function``
        ``orignal.__doc__`` is appended to the ``__doc__`` of the ``target``

    Examples
    ========

    >>> def ho():
    ...     '''Ho'''
    ...     print(ho.__doc__)
    >>> @append_doc(ho)
    ... def hi():
    ...     '''Hi'''
    ...     print(hi.__doc__)
    >>> hi()
    HiHo
    """
    return _append(original.__doc__)
    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        target.__doc__ += original.__doc__
        return target
    return wrapper

def _insert(txt):
    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        target.__doc__ = txt + target.__doc__
        return target
    return wrapper

def insert_str(txt):
    return _insert(txt)

def insert_doc(original):
    """
    Inserts the docstring from passed function ``original`` in the ``target`` function docstring.
    """
    return _insert(original.__doc__)
    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        target.__doc__ = original.__doc__ + target.__doc__
        return target
    return wrapper

def insert_eq():
    """
    Inserts the function and its parameters and an equal sign
    """
    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        safe = target.__doc__
        target.__doc__ = target.__name__  + "("
        for v in target.__code__.co_varnames:
            target.__doc__ += v + ","
        target.__doc__ = target.__doc__[:-1]
        target.__doc__ += ") = " + safe
        return target
    return wrapper

def insert_latex():
    """
    TODO 
    """
    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        safe = target.__doc__
        target.__doc__ = target.__name__  + "("
        for v in target.__code__.co_varnames:
            target.__doc__ += v + ","
        target.__doc__ = target.__doc__[:-1]
        target.__doc__ += ") = " + safe
        return target
    return wrapper


def table_sep(tabs=1):
    return "==============  ==============  ==============\n" +"\t".join(["" for i in range(0,tabs+1)])
def table(dic,top=True,bottom=True,init=True,tabs=1):
    """
    Add dict= {'key': [values...]} to a simple reST table.
    """
    t =  util.times("\t",tabs)
    rs = ""
    if init:
        rs += t
    if top:
        rs += table_sep(tabs=tabs)
    for k,vs in dic.items():
        rs += str(k) 
        p = str(k)
        for v in vs:
            rs += "".join([' ' for i in range(16-len(p))]) + str(v)
            p = str(v)
        rs += "\n" + t
    if bottom:
        rs += table_sep(tabs=tabs)
    return rs


if __name__ == "__main__":
    import doctest
    doctest.testmod()